"""Reasoner engine."""
import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

from jinja2 import Template
from pydantic import BaseModel

from ontogpt.engines.knowledge_engine import KnowledgeEngine
from ontogpt.ontex.extractor import (
    Answer,
    Axiom,
    Explanation,
    LLMReasonMethodType,
    Task,
    TaskCollection,
)
from ontogpt.prompts.reasoning import DEFAULT_REASONING_PROMPT
from ontogpt.utils.parse_utils import split_on_one_of

logger = logging.getLogger(__name__)


def flatten_list(lst):
    flat_list = []
    for item in lst:
        if isinstance(item, list):
            flat_list.extend(flatten_list(item))
        else:
            flat_list.append(item)
    return flat_list


class ReasonerResult(BaseModel):
    """The result of a reason query."""

    name: Optional[str] = None
    completed: Optional[bool] = True
    task_name: Optional[str] = None
    task_type: Optional[str] = None
    task_obfuscated: Optional[bool] = None
    method: Optional[LLMReasonMethodType] = None
    model: Optional[str] = None
    description: Optional[str] = None
    answers: Optional[List[Answer]] = None
    prompt: Optional[str] = None
    completion: Optional[str] = None
    jaccard_score: Optional[float] = 0.0
    false_positives: Optional[List[str]] = None
    false_negatives: Optional[List[str]] = None
    num_false_positives: Optional[int] = None
    num_false_negatives: Optional[int] = None
    num_true_positives: Optional[int] = None
    num_true_negatives: Optional[int] = None
    precision: Optional[float] = 0.0
    recall: Optional[float] = None
    f1_score: Optional[float] = None
    len_shortest_explanation: Optional[int] = None

    class Config:
        """Pydantic config."""

        use_enum_values = True


class ReasonerResultSet(BaseModel):
    name: str = ""
    results: List[ReasonerResult]


@dataclass
class ReasonerEngine(KnowledgeEngine):
    """Engine for performing reasoning using GPT.

    This engine takes as input an Ontology, and a query Task,
    and then translates this to a GPT prompt that asks GPT to
    perform the task over the ontology after reasoning over it.

    The Task is typically a query such as finding superclasses of
    a given class.

    This is intended primarily for investigation purposes. For practical
    scenarios, it is recommended to use a dedicated OWL reasoner. The goal
    of this engine is to evaluate the extent to which GPT can perform
    reasoning-like tasks, including deduction and abduction (explanation).

    Due to token-length constraints on GPT models, it is usually necessary
    to extract a submodule prior to reasoning. This can be done using the
    OntologyExtractor:

    >>> from oaklib import get_adapter
    >>> from ontogpt.ontex.extractor import OntologyExtractor, Task
    >>> adapter = get_adapter("sqlite:obo:go")
    >>> extractor = OntologyExtractor(adapter=adapter)
    >>> task = extractor.extract_indirect_superclasses_task(
    ...    subclass="GO:0005634", siblings=["GO:0005773"], roots=["GO:0043226"]
    ... )

    The extractor will actually perform the intended reasoning task, and include
    the expected answer in the Task object. This may seem a little circular, but
    the goal here is to evaluate.

    The task can then be passed to the reasoner engine:

    >>> from ontogpt.engines.reasoner_engine import ReasonerEngine
    >>> reasoner = ReasonerEngine()
    >>> result = reasoner.reason(task)

    We can expand on this:

    >>> for answer in result.answers:
    ...    print(f"PARENT: {answer.text}")
    <BLANKLINE>
    PARENT: MembraneBoundedOrganelle
    ...
    PARENT: IntracellularOrganelle

    We can also ask for explanations:

    >>> task.include_explanations = True
    >>> result = reasoner.reason(task)
    >>> for answer in result.answers:
    ...    print(f"PARENT: {answer.text}")
    ...    if answer.explanations:
    ...        print("  EXPLANATION:")
    ...        for e in answer.explanations:
    ...            for axiom in e.axioms:
    ...                print(f"    AXIOM: {axiom.text}")
    <BLANKLINE>
    ...
    PARENT: IntracellularOrganelle
        EXPLANATION:
            AXIOM: Nucleus SubClassOf IntracellularMembraneBoundedOrganelle
            AXIOM: IntracellularMembraneBoundedOrganelle SubClassOf IntracellularOrganelle

    """

    completion_length = 250

    def reason(
        self, task: Task, template_path=None, strict=False, evaluate: bool = None
    ) -> ReasonerResult:
        """Reason over axioms and query entailments."""
        if template_path is None:
            template_path = DEFAULT_REASONING_PROMPT
        if isinstance(template_path, Path):
            template_path = str(template_path)
        if isinstance(template_path, str):
            # create a Jinja2 template object
            with open(template_path) as file:
                template_txt = file.read()
                template = Template(template_txt)
        prompt = template.render(
            task=task,
            ontology=task.ontology,
            query=task.query,
            examples=task.examples,
        )
        completion_length = self.completion_length
        if task.method == LLMReasonMethodType.EXPLANATION:
            completion_length *= 2
        elif task.method == LLMReasonMethodType.CHAIN_OF_THOUGHT:
            completion_length *= 2
        logger.info(f"Prompt: {prompt}")
        if self.encoding is not None:
            prompt_length = len(self.encoding.encode(prompt)) + 10
        else:
            prompt_length = len(prompt)
        completed = True
        logger.info(f"PROMPT LENGTH: {prompt_length}")

        payload = self.client.complete(prompt, max_tokens=completion_length)
        if task.has_multiple_answers:
            elements = payload.split("- ")
            answers = [self._parse_single_answer(e, task) for e in elements]
        else:
            answers = [self._parse_single_answer(payload, task)]
        answers = [a for a in flatten_list(answers) if a is not None]
        result = ReasonerResult(
            completed=completed,
            task_name=task.name,
            task_type=task.type,
            task_obfuscated=task.obfuscated,
            method=task.method,
            len_shortest_explanation=task.len_shortest_explanation,
            model=self.model,
            prompt=prompt,
            completion=payload,
        )
        # TODO: determine which it doesn't work to initialize with this
        result.answers = answers  # type: ignore
        logger.debug(f"Answers: {task.answers} // {answers}")
        result.name = f"{task.name}-{task.method.value}-{self.model}"
        if not task.answers and evaluate:
            raise ValueError(f"Cannot evaluate without expected answers: {task}")
        if task.answers is not None:
            self.evaluate(result, task)
        return result

    def reason_multiple(self, task_collection: TaskCollection, **kwargs) -> ReasonerResultSet:
        """
        Reason over multiple tasks.

        :param task_collection:
        :param kwargs:
        :return:
        """
        results = [self.reason(task, **kwargs) for task in task_collection.tasks]
        return ReasonerResultSet(results=results)

    def _parse_single_answer(
        self, payload: str, task: Task
    ) -> Optional[Union[Answer, List[Answer]]]:
        """Parse single answer from payload.

        In some cases for COT reasoning the model may compress multiple
        answers into a single answer line.
        """
        payload = payload.strip()
        if not payload:
            return None
        payload = payload.replace("\n", " ")
        if task.chain_of_thought:
            pattern = r"^REASONING:\s*\[\s*(.*?)\s*\]\s*CONCLUSION:\s*(.*?)$"
            match = re.match(pattern, payload)
            print(f"CHAIN OF THOUGHT: {payload}")
            if match:
                print(f"MATCH: {match.groups()}")
                explanation = match.group(1)
                text = match.group(2)
                # rest = match.group(3)
                axioms = [x.strip() for x in split_on_one_of(explanation, [";", ","])]
                explanation = Explanation(axioms=[Axiom(text=a) for a in axioms if a])
                axiom_texts = split_on_one_of(text, [";", ","])
                if len(axiom_texts) == 1:
                    return Answer(text=text, explanations=[explanation])
                else:
                    return [
                        Answer(text=axiom_text, explanations=[explanation])
                        for axiom_text in axiom_texts
                    ]
        pattern = r"^(.*?)\s*\[\s*(.*?)\s*\]\s*(.*?)$"
        match = re.match(pattern, payload)
        if match:
            text = match.group(1)
            explanation = match.group(2)
            rest = match.group(3)
            axioms = [x.strip() for x in split_on_one_of(explanation, [";", ","])]
            explanation = Explanation(axioms=[Axiom(text=a) for a in axioms if a])
            if rest:
                explanation.comments = [rest]
            return Answer(text=text, explanations=[explanation])
        else:
            return Answer(text=payload)

    def evaluate(self, result: ReasonerResult, task: Task):
        """Evaluate result against task."""
        logger.debug(f"Evaluating result: {result}")
        positives = {t.text for t in task.answers}
        result_answer_texts = {a.text for a in result.answers}
        ixn = positives.intersection(result_answer_texts)
        all_texts = positives.union(result_answer_texts)
        result.false_positives = list(result_answer_texts - positives)
        result.false_negatives = list(positives - result_answer_texts)
        result.num_false_positives = len(result.false_positives)
        result.num_false_negatives = len(result.false_negatives)
        result.num_true_positives = len(ixn)
        tp_plus_tn = result.num_true_positives + result.num_false_positives
        if tp_plus_tn == 0:
            result.precision = 0.0
        else:
            result.precision = result.num_true_positives / tp_plus_tn
        result.recall = result.num_true_positives / len(positives)
        if len(all_texts) == 0:
            result.jaccard_score = 0.0
        else:
            result.jaccard_score = len(ixn) / len(all_texts)
        if result.num_true_positives == 0:
            result.f1_score = 0.0
        else:
            result.f1_score = (
                2 * (result.precision * result.recall) / (result.precision + result.recall)
            )



"""Utilities for retrieving and applying prebuild models."""

import logging
from pathlib import PosixPath

import pystow

ONTOGPT_MODULE = pystow.module("ontogpt")


def get_model(url: str) -> PosixPath:
    """Retrieve a model from a given URL.

    Returns the Path for the retrieved file,
    or the path to where it already exists.
    """
    logging.info(f"Retrieving model from {url} if needed...")
    mod_path = ONTOGPT_MODULE.ensure(url=url, force=False, download_kwargs={"backend": "requests"})
    logging.info(f"Model now at {mod_path}")

    return mod_path



"""Similarity engine."""

import logging
from dataclasses import dataclass
from typing import Iterable, List

from oaklib import BasicOntologyInterface
from oaklib.datamodels.vocabulary import IS_A

from ontogpt.clients import LLMClient
from ontogpt.engines.knowledge_engine import KnowledgeEngine

logger = logging.getLogger(__name__)


@dataclass
class EmbeddingSimilarity:
    subject_id: str = ""
    subject_label: str = ""
    object_id: str = ""
    object_label: str = ""
    embedding_cosine_similarity: float = None
    object_rank_for_subject: int = None


@dataclass
class SimilarityEngine(KnowledgeEngine):
    """Engine for generating synonyms."""

    adapter: BasicOntologyInterface = None
    autolabel: bool = True
    definitions: bool = True
    parents: bool = True
    ancestors: bool = True
    synonyms: bool = True
    logical_definitions: bool = False

    def similarity(self, entity1: str, entity2: str) -> EmbeddingSimilarity:
        """Get similarity."""
        t1 = self.entity_text(entity1)
        t2 = self.entity_text(entity2)
        client = LLMClient()
        score = client.similarity(t1, t2)
        obj = EmbeddingSimilarity(
            subject_id=entity1, object_id=entity2, embedding_cosine_similarity=score
        )
        if self.autolabel:
            obj.subject_label = self.adapter.label(entity1)
            obj.object_label = self.adapter.label(entity2)
        return obj

    def search(self, entity1: str, entities: List[str]) -> Iterable[EmbeddingSimilarity]:
        """Get similarity."""
        client = LLMClient()
        t1 = self.entity_text(entity1)
        sims = []
        for entity2 in entities:
            t2 = self.entity_text(entity2)
            score = client.similarity(t1, t2)
            sim = EmbeddingSimilarity(
                subject_id=entity1, object_id=entity2, embedding_cosine_similarity=score
            )
            if self.autolabel:
                sim.subject_label = self.adapter.label(entity1)
                sim.object_label = self.adapter.label(entity2)
            sims.append(sim)
        sims = sorted(sims, key=lambda x: x.embedding_cosine_similarity, reverse=True)
        for i, sim in enumerate(sims):
            sim.object_rank_for_subject = i
        yield from sims

    def entity_text(self, entity: str) -> str:
        """Get text for an entity."""
        adapter = self.adapter
        s = f"{adapter.label(entity)}"
        if self.definitions:
            s += f"\ndefinition: {adapter.definition(entity)}"
        if self.parents:
            parent_labels = [
                adapter.label(o) for _s, _p, o in adapter.relationships([entity], [IS_A])
            ]
            s += f"\nparents: {'; '.join(parent_labels)}"
        if self.ancestors:
            ancestor_labels = [
                adapter.label(a) for a in adapter.ancestors([entity], [IS_A], reflexive=False)
            ]
            s += f"\nancestors: {'; '.join(ancestor_labels)}"
        if self.synonyms:
            s += f"\nsynonyms: {'; '.join(adapter.entity_aliases(entity))}"
        if self.logical_definitions:
            for ldef in adapter.logical_definitions(entity):
                genus_labels = [adapter.label(g) for g in ldef.genusIds]
                restriction_labels = [
                    f"{adapter.label(r.propertyId)} {adapter.label(r.valueId)}"
                    for r in ldef.restrictions
                ]
                s += f"\nlogical definition: A {', '.join(genus_labels)} that\
                    {' and '.join(restriction_labels)}"
            s += f"\nlogical definitions: {'; '.join(adapter.logical_definitions(entity))}"
        logger.info(f"Entity text for {entity}: {s}")
        return s



"""Output parser utilities for bulleted output."""

import time
from oaklib import get_adapter

NULL_VALS = [
    "",
    "Not mentioned",
    "none mentioned",
    "Not mentioned in the text",
    "Not mentioned in the provided text.",
    "No exposures mentioned in the text.",
    "None",
    "N/A",
    "No exposures mentioned in the text.",
    "None mentioned in the text",
    "Not mentioned in the text.",
    "None relevant",
    "None mentioned in the text.",
    "No gene to molecular activity relationships mentioned in the text.",
    "No genes mentioned",
    "No genes mentioned in the text.",
]
lines = []
adapter = get_adapter("sqlite:obo:MONDO")
# output_file = "output_2000_0719.yaml"
# output_file = "output_100_0817.yaml"
# output_file = "output_100_0818.yaml"
output_file = "output_100.0911.yaml"


def tripleprint(dict):
    key = next(iter(dict))
    for i in range(len(dict[key])):
        curr = []
        for value in dict.values():
            curr.append(value[i])
        print(curr)


def forprint(lst):
    for elem in lst:
        print(elem)


def forprint_newline(lst):
    for elem in lst:
        print(elem)
        print("\n")


def nprint(lst, n):
    for i in range(n):
        print(lst[i])


def sleepprint(lst):
    for elem in lst:
        print(elem)
        time.sleep(0.5)


def enumprint(lst):
    for index, elem in enumerate(lst):
        print(str(index + 1) + ":\t" + str(elem))


with open(output_file, "r") as file:
    to_print = False
    # perpetuators = tuple(["  subject:", "  predicate:", "  object:", "    "])
    perpetuators = tuple(["    - subject:", "      predicate:", "      object:"])
    perpetuators = tuple(
        [
            "    - subject:",
            "      predicate:",
            "      object:",
            "      qualifier:",
            "      subject_qualifier:",
            "      object_qualifier:",
        ]
    )
    for line in file:
        line = line.strip("\n")
        # if line.startswith("extracted_object:"):
        if line.startswith("  disease_cellular_process_relationships:") and not line.startswith(
            "  disease_cellular_process_relationships: "
        ):
            # if line.startswith(perpetuators):
            to_print = True
        # elif not line.startswith(perpetuators):
        # if line.startswith("named_entities:"):
        elif not line.startswith(perpetuators):
            to_print = False
        if to_print:
            lines.append(line)

cleaned_lines = [x for x in list(filter(lambda elem: not (elem.isspace()), lines)) if x.strip()]
cleaned_lines = [
    x
    for x in cleaned_lines
    if not x.startswith(
        tuple(
            [
                "      qualifier:",
                "      subject_qualifier:",
                "      object_qualifier:",
                "  disease_cellular_process_relationships:",
            ]
        )
    )
]

header = "    - subject:"
i = 0
while i < len(cleaned_lines):
    if cleaned_lines[i].startswith(header):
        for index, elem in enumerate(cleaned_lines[i + 1 : i + 3]):
            if elem.startswith(header):
                next_index = i + 1 + index
                del cleaned_lines[i:next_index]
                i -= 1
    i += 1

cleaned_lines = [x.strip() for x in cleaned_lines]
for i, line in enumerate(cleaned_lines.copy()):
    # for line in cleaned_lines:
    if line.startswith("- "):
        cleaned_lines[i] = line[2:]

grouped_lines = [cleaned_lines[n : n + 3] for n in range(0, len(cleaned_lines), 3)]
trimmed_dict: dict = {"diseases": [], "relationships": [], "cellular processes": []}
for group in grouped_lines:
    subject = group[0].split(":", 1)[1].strip()
    relation = group[1].split(":", 1)[1].strip()
    object = group[2].split(":", 1)[1].strip()
    trimmed_dict["diseases"].append(subject)
    trimmed_dict["relationships"].append(relation)
    trimmed_dict["cellular processes"].append(object)

for _key, value in trimmed_dict.copy().items():
    for i, elem in enumerate(value):
        if elem.lower() in (val.lower() for val in NULL_VALS):
            for _key, value in trimmed_dict.items():
                del value[i]

# subjects_with_chebi = []
# objects_with_ecto = []
subjects_with_mondo: list = []
mondo_go = []
key = next(iter(trimmed_dict))
for i in range(len(trimmed_dict[key])):
    curr = []
    for value in trimmed_dict.values():
        curr.append(value[i])
    # subjects_with_chebi.append(curr)
    mondo_go.append(curr)
    # objects_with_ecto.append(curr)
# subjects_with_chebi = [x for x in subjects_with_chebi if x[0].startswith("CHEBI:")]
mondo_go = [x for x in mondo_go if x[0].startswith("MONDO:") and x[2].startswith("GO:")]
# objects_with_ecto = [x for x in objects_with_ecto if x[2].startswith("ECTO:")]

forprint(mondo_go)

# forprint(subjects_with_mondo)

subjects_with_names = []
# for elem in subjects_with_chebi:
for elem in subjects_with_mondo:
    curr = [elem[:]]
    curr[0] = adapter.label(curr[0])
    subjects_with_names.append(curr)

# forprint(subjects_with_names)



from __future__ import annotations 
from datetime import (
    datetime,
    date
)
from decimal import Decimal 
from enum import Enum 
import re
import sys
from typing import (
    Any,
    List,
    Literal,
    Dict,
    Optional,
    Union
)
from pydantic.version import VERSION  as PYDANTIC_VERSION 
if int(PYDANTIC_VERSION[0])>=2:
    from pydantic import (
        BaseModel,
        ConfigDict,
        Field,
        field_validator
    )
else:
    from pydantic import (
        BaseModel,
        Field,
        validator
    )

metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment = True,
        validate_default = True,
        extra = "forbid",
        arbitrary_types_allowed = True,
        use_enum_values = True,
        strict = False,
    )
    pass


class NullDataOptions(str, Enum):
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    NOT_MENTIONED = "NOT_MENTIONED"


class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")


class NamedEntity(ConfiguredBaseModel):
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class CompoundExpression(ConfiguredBaseModel):
    pass


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")


class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)


class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    publication: Optional[Publication] = Field(None)
    entities: Optional[List[str]] = Field(default_factory=list)


class RelationshipType(NamedEntity):
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Publication(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")


class AnnotatorResult(ConfiguredBaseModel):
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)


class ValidationReport(NamedEntity):
    """
    A report object containing one or more validation errors or notifications.
    """
    fatal_result_summary: Optional[List[str]] = Field(default_factory=list, description="""Identify similarities among errors in the report with the category of FATAL. What is similar about these errors? Do they include many of the same prefixes or namespaces? If the same patterns or rules are violated, what are they? Separate each summary with a semicolons. If no errors are present, this field should be NONE.""")
    error_result_summary: Optional[List[str]] = Field(default_factory=list, description="""Identify similarities among errors in the report with the category of ERROR. What is similar about these errors? Do they include many of the same prefixes or namespaces? If the same patterns or rules are violated, what are they? Separate each summary with a semicolons. If no errors are present, this field should be NONE.""")
    warning_result_summary: Optional[List[str]] = Field(default_factory=list, description="""Identify similarities among errors in the report with the category of WARNING. What is similar about these errors? Do they include many of the same prefixes or namespaces? If the same patterns or rules are violated, what are they? Separate each summary with a semicolons. If no errors are present, this field should be NONE.""")
    info_result_summary: Optional[List[str]] = Field(default_factory=list, description="""Identify similarities among errors in the report with the category of INFO. What is similar about these errors? Do they include many of the same prefixes or namespaces? If the same patterns or rules are violated, what are they? Separate each summary with a semicolons. If no errors are present, this field should be NONE.""")
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
ValidationReport.model_rebuild()



"""Client for running LLM completion requests through LiteLLM."""

import litellm
import logging
from dataclasses import dataclass, field

from litellm import completion, embedding
from litellm.caching import Cache
import numpy as np
from oaklib.utilities.apikey_manager import get_apikey_value
import openai  # For error handling

from ontogpt import DEFAULT_MODEL, MODELS

logger = logging.getLogger(__name__)

# Just get the part before the slash in each model name
SERVICES = {model.split("/")[0] for model in MODELS.keys() if len(model.split("/")) > 1}

@dataclass
class LLMClient:

    model: str = field(default_factory=lambda: DEFAULT_MODEL)
    cache_db_path: str = ""
    api_key: str = ""
    api_base: str = None
    api_version: str = None

    # litellm uses this param to specify client when it isn't clear from
    # the model name
    custom_llm_provider: str = None

    temperature: float = 1.0

    def __post_init__(self):
        # Get appropriate API key for the model source
        # and other details if needed
        if not self.api_key and not self.custom_llm_provider:
            self.api_key = get_apikey_value("openai")
        elif self.custom_llm_provider == "anthropic":
            self.api_key = get_apikey_value("anthropic-key")
        else:
            for service in SERVICES:
                if self.model.startswith(service):
                    self.api_key = get_apikey_value(service + "-key")
                    if service == "azure" and self.api_base is None:
                        self.api_base = get_apikey_value(service + "-base")
                    if service == "azure" and self.api_version is None:
                        self.api_version = get_apikey_value(service + "-version")

        # Set up the cache, and set the cache path if provided
        if len(self.cache_db_path) == 0:
            litellm.cache = Cache(type="disk")
        else:
            litellm.cache = Cache(type="disk", disk_cache_dir=self.cache_db_path)

    def complete(self, prompt, show_prompt: bool = False, **kwargs) -> str:

        logger.info(f"Complete: engine={self.model}, prompt[{len(prompt)}]={prompt[0:100]}...")
        if show_prompt:
            logger.info(f" SENDING PROMPT:\n{prompt}")

        response = None

        try:
            # TODO: expose user prompt to CLI
            response = completion(
                api_key=self.api_key,
                api_base=self.api_base,
                api_version=self.api_version,
                model=self.model,
                messages=[{"content": prompt, "role": "user"}],
                temperature=self.temperature,
                caching=True,
                custom_llm_provider=self.custom_llm_provider,
            )
        except openai.APITimeoutError as e:
            print(f"Encountered API timeout error: {e}")
        except Exception as e:
            print(f"Encountered error: {type(e)}, Error: {e}")

        if response is not None:
            payload = response.choices[0].message.content
        else:
            logger.error("No response or response is empty.")
            payload = ""

        return payload

    def embeddings(self, text: str):
        text = str(text)

        # TODO: set embedding model based on model source
        # Or at least set the default for OpenAI models
        if self.model is None:
            model = "text-embedding-ada-002"
        else:
            model = self.model

        logger.info(f"Retrieving embeddings from {model} for text: {text[0:80]}...")

        response = embedding(
            api_key=self.api_key,
            api_base=self.api_base,
            api_version=self.api_version,
            model=model,
            input=[text],
            caching=True,
        )

        if response is not None:
            payload = response.data[0]['embedding']
        else:
            logger.error("No response or response is empty.")
            payload = ""

        return payload

    def similarity(self, text1: str, text2: str, **kwargs):
        a1 = self.embeddings(text1, **kwargs)
        a2 = self.embeddings(text2, **kwargs)
        logger.debug(f"similarity: {a1[0:10]}... x {a2[0:10]}... // ({len(a1)} x {len(a2)})")
        return np.dot(a1, a2) / (np.linalg.norm(a1) * np.linalg.norm(a2))

    def euclidian_distance(self, text1: str, text2: str, **kwargs):
        a1 = self.embeddings(text1, **kwargs)
        a2 = self.embeddings(text2, **kwargs)
        return np.linalg.norm(np.array(a1) - np.array(a2))



"""Utilities for parsing text."""

import re

from typing import List


def split_on_one_of(text: str, separators: List[str]) -> List[str]:
    """Split text on the first separator found."""
    for sep in separators:
        if sep in text:
            return text.split(sep)
    return [text]

def get_section_of_interest(data, tag_of_interest):
    # I blame adobe
    # Find the index of the element that matches the case-insensitive regex pattern
    start_index = None
    pattern = re.compile(tag_of_interest, re.IGNORECASE)
    if isinstance(data, str):
        data = data.split("\n")
    for i, item in enumerate(data):
        if pattern.search(item):
            start_index = i
            break

    if start_index is not None:
        # Find the index of the next element that starts with '<p>'
        next_index = next(
            (
                i
                for i, item in enumerate(data[start_index + 1 :], start=start_index + 1)
                if item.startswith("<p>")
            ),
            None,
        )

        if next_index is not None:
            # Extract the desired element
            result = data[next_index]
            return result
        else:
            raise ValueError("No element starting with '<p>' found after the tag_of_interest")
    else:
        raise ValueError("No element matching the tag_of_interest found in the list.")


"""Wikipedia Client."""

from dataclasses import dataclass

import requests
import wikipediaapi


@dataclass
class WikipediaClient:
    """A client for extracting text from Wikipedia articles."""

    language: str = "en"

    def text(self, title: str) -> str:
        """Get the text of an article."""
        wiki = wikipediaapi.Wikipedia("ontogpt", self.language)
        page = wiki.page(title)

        if page.exists():
            return page.text
        else:
            raise ValueError(f"Page {title} does not exist.")

    def search_wikipedia_articles(self, topic, results=10):
        base_url = f"https://{self.language}.wikipedia.org/w/api.php"
        params = {
            "action": "query",
            "list": "search",
            "format": "json",
            "srsearch": topic,
            "srprop": "size|wordcount",
            "srlimit": results,
            "srinfo": "totalhits",
        }
        response = requests.get(base_url, params=params)
        data = response.json()

        if data and "query" in data and "search" in data["query"]:
            return data["query"]["search"]
        else:
            return []



"""ontogpt package."""

from importlib import metadata
from pathlib import Path

from litellm import get_model_cost_map

# Define the default model
rel_path = Path(__file__).resolve()

DEFAULT_MODEL = "gpt-4o"

# Build list of available models
MODELS = get_model_cost_map("")

try:
    __version__ = metadata.version(__name__)
except metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"  # pragma: no cover



"""Webapp main function."""

from io import StringIO
from pathlib import Path
from typing import Dict

import uvicorn
from fastapi import FastAPI, Form, Request
from pydantic import BaseModel
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates

from ontogpt.cli import _get_templates
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.io.html_exporter import HTMLExporter
from ontogpt.io.template_loader import get_template_details

this_path = Path(__file__).parent
static_dir = this_path / "static"
html_dir = this_path / "html"

# Populate the dict of available models
DATAMODELS = {}
all_templates = _get_templates()
all_templates = dict(sorted(all_templates.items()))
for template_id, (_name, description) in all_templates.items():
    DATAMODELS[template_id] = f"{description}"

LLM_MODELS = ["gpt-4o", "gpt-4", "gpt-4-turbo", "gpt-3.5-turbo"]


class Query(BaseModel):
    text: str
    datamodel: str
    model: str


app = FastAPI()
static_dir.mkdir(parents=True, exist_ok=True)
app.mount("/static", StaticFiles(directory=str(static_dir)), name="static")
templates = Jinja2Templates(directory=str(html_dir))

html_exporter = HTMLExporter(output=None)
engines: Dict[str, SPIRESEngine] = {}


def get_engine(datamodel: str, llm_model: str):
    if datamodel not in engines:
        template_details = get_template_details(template=datamodel)
        try:
            engines[datamodel] = SPIRESEngine(
                model=llm_model, template_details=template_details, model_provider="openai"
            )
        except ValueError as e:
            print(f"Encountered an error setting up the knowledge engine: {e}")
            print("Will fall back to defaults.")
            engines[datamodel] = SPIRESEngine(
                model="gpt-3.5-turbo", template_details=template_details, model_provider="openai"
            )
    return engines[datamodel]


@app.get("/")
def read_root(request: Request):
    return templates.TemplateResponse(
        "form.html",
        context={"request": request, "datamodels": DATAMODELS, "llm_models": LLM_MODELS},
    )


@app.post("/")
def form_post(
    request: Request, datamodel: str = Form(...), text: str = Form(...), llm_model: str = Form(...)
):
    print(f"Received request with schema {datamodel}")
    print(f"Received request with text {text}")
    print(f"Received request to be sent to {llm_model}")
    engine = get_engine(datamodel, llm_model)
    ann = engine.extract_from_text(text)
    print(f"Got {ann}")
    output = StringIO()
    html_exporter.export(extraction_output=ann, output=output)
    return templates.TemplateResponse(
        "results.html", context={"request": request, "inner_html": output.getvalue()}
    )


def start():
    """Launch with `poetry run start` at root level."""
    uvicorn.run("ontogpt.webapp.main:app", host="127.0.0.1", port=8000, reload=True)



"""Resolver engine."""

from typing import Optional, Type, Union

from class_resolver import ClassResolver

from ontogpt.engines.halo_engine import HALOEngine  # type: ignore
from ontogpt.engines.knowledge_engine import KnowledgeEngine
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.io.template_loader import get_template_details

resolver = ClassResolver([SPIRESEngine, HALOEngine], base=KnowledgeEngine)


def create_engine(
    template: str, engine: Optional[Union[str, Type, KnowledgeEngine]] = None, **kwargs
) -> Union[KnowledgeEngine, SPIRESEngine]:
    """Create a knowledge engine."""
    template_details = get_template_details(template=template)
    if engine is None:
        engine = SPIRESEngine(template_details=template_details)
    if isinstance(engine, str):
        engine = resolver.get_class(engine)(**kwargs)
    if engine is not None and not isinstance(engine, str):
        return engine(template_details=template_details, **kwargs)  # type: ignore
    else:
        return SPIRESEngine  # type: ignore



"""Utility for running multilingual analysis."""

import codecs
import logging
import os
from io import TextIOWrapper


from ontogpt.clients import LLMClient
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.io.template_loader import get_template_details
from ontogpt.io.yaml_wrapper import dump_minimal_yaml


def multilingual_analysis(input_data_dir, output_directory, output, model="gpt-4-turbo"):
    """Run the multilingual analysis."""
    # Set up the extraction template
    template = "all_disease_grounding"
    template_details = get_template_details(template=template)

    # make sure the output directory exists
    os.makedirs(output_directory, exist_ok=True)

    # Set up the writer object
    if not isinstance(output, TextIOWrapper):
        output = codecs.getwriter("utf-8")(output)

    ai = LLMClient()
    ai.model = model

    # Keep track of the predictions
    # Key is the filename, value is a list of predictions
    pred_ids = {}
    pred_names = {}

    # Log all errors, with prompt filename as key and error as value
    errors = {}

    for filename in os.listdir(input_data_dir):
        completed = False
        grounded = False
        if filename.endswith(".txt"):
            file_path = os.path.join(input_data_dir, filename)

            with open(file_path, mode="r", encoding="utf-8") as txt_file:
                prompt = txt_file.read()

            try:
                gpt_diagnosis = ai.complete(prompt)
                completed = True
            except Exception as e:
                errors[filename] = e
                logging.error(f"Error: {e}")

            # Call the extract function here
            # to ground the answer to OMIM (using MONDO, etc)
            # The KE is refreshed here to avoid retaining
            if completed:
                try:
                    ke = SPIRESEngine(
                        template_details=template_details,
                        model=model,
                    )
                    extraction = ke.extract_from_text(text=gpt_diagnosis)
                    predictions = extraction.named_entities
                    pred_ids[filename] = []
                    pred_names[filename] = []
                    for pred in predictions:
                        pred_ids[filename].append(pred.id)
                        pred_names[filename].append(pred.label)

                    # Log the result
                    logging.info(
                        "input file name" "\tpredicted diagnosis ids\tpredicted diagnosis names\n"
                    )
                    logging.info(
                        f"{filename}"
                        f'\t{"|".join(pred_ids[filename])}'
                        f'\t{"|".join(pred_names[filename])}\n'
                    )
                    grounded = True
                except Exception as e:
                    errors[filename] = e
                    logging.error(f"Error: {e}")

            # Retain the output as text
            # Create the output filename based on the input filename
            output_file_name = filename + ".result"
            output_file_path = os.path.join(output_directory, output_file_name)
            with open(output_file_path, "w", encoding="utf-8") as outfile:
                if completed and grounded:
                    outfile.write(gpt_diagnosis)

                    # Write the result
                    # Include the input filename for the prompt in the output
                    extraction.extracted_object.label = filename
                    output.write("---\n")
                    output.write(dump_minimal_yaml(extraction))

                else:
                    outfile.write(f"Error: {errors[filename]}")

            # If there were errors, log them to a file
            if len(errors) > 0:
                error_file_path = os.path.join(output_directory, "errors.txt")
                with open(error_file_path, "w", encoding="utf-8") as outfile:
                    for error in errors:
                        outfile.write(f"{error}\t{errors[error]}\n")



"""Ontology converter."""

import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional

from linkml_runtime import SchemaView
from linkml_runtime.utils.formatutils import camelcase
from oaklib.datamodels.obograph import Graph
from oaklib.datamodels.vocabulary import IS_A
from oaklib.interfaces.obograph_interface import OboGraphInterface

from ontogpt.templates.halo import Ontology, OntologyElement

this_path = Path(__file__).parent
logger = logging.getLogger(__name__)


@dataclass
class OntologyConverter:
    """Converts an OAK ontology to an OntoGPT schema."""

    adapter: OboGraphInterface = None
    schemaview: SchemaView = None
    fixed_slot_values: Dict[str, str] = field(default_factory=lambda: {})

    def __post_init__(self):
        templates_path = this_path.parent / "templates"
        path_to_template = str(templates_path / "halo.yaml")
        self.schemaview = SchemaView(path_to_template)

    def extract_seed_ontology(self, seeds: List[str], predicates: List[str]) -> Ontology:
        """Extract an ontology from a given text.

        :param text:
        :return:
        """
        ancestors = list(set(list(self.adapter.ancestors(seeds, predicates, reflexive=True))))
        seed_graph = self.adapter.extract_graph(ancestors, predicates, dangling=False)
        logger.info(len(seed_graph.nodes))
        seed_ontology = self.from_obograph(seed_graph)
        return seed_ontology

    def from_adapter(self) -> Ontology:
        """Convert an OAK adapter to an Ontology.

        :param adapter:
        :return:
        """
        graph = self.adapter.as_obograph()
        return self.from_obograph(graph)

    def from_obograph(self, graph: Graph) -> Ontology:
        """Convert an OBO Graph to an Ontology.

        :param graph:
        :return:
        """
        adapter = self.adapter
        ontology = Ontology()
        element_index = {}
        node_to_element_name = {}
        id2slot = {}
        inverses = {}
        for slot in self.schemaview.class_induced_slots(OntologyElement.__name__):
            if slot.inverse:
                inverses[slot.name] = slot.inverse
                inverses[slot.inverse] = slot.name
            if slot.slot_uri:
                id2slot[slot.slot_uri] = slot
        logger.info(list(id2slot.keys()))
        logger.info(inverses)
        for node in graph.nodes:
            meta = node.meta
            if not node.lbl:
                continue
            if not meta:
                # logger.warning(f"Node {node.id} has no meta")
                continue
            element = OntologyElement(
                name=self.node_to_name(node.id, node.lbl),
                synonyms=[synonym.val for synonym in meta.synonyms],
                description=meta.definition.val if meta.definition else None,
            )
            for k, v in self.fixed_slot_values.items():
                setattr(element, k, v)
            element_index[element.name] = element
            node_to_element_name[node.id] = element.name
        for edge in graph.edges:
            if edge.pred == "is_a":
                pred = IS_A
            else:
                try:
                    pred = adapter.uri_to_curie(edge.pred)
                except:
                    pred = edge.pred
            if pred not in id2slot:
                continue
            if edge.sub not in node_to_element_name:
                continue
            if edge.obj not in node_to_element_name:
                continue
            subject = node_to_element_name[edge.sub]
            object = node_to_element_name[edge.obj]
            slot = id2slot[pred]
            getattr(element_index[subject], slot.name).append(object)
            if slot.name in inverses:
                inverse = inverses[slot.name]
                getattr(element_index[object], inverse).append(subject)
        for ldef in adapter.logical_definitions([node.id for node in graph.nodes]):
            if ldef.definedClassId in node_to_element_name:
                element = element_index[node_to_element_name[ldef.definedClassId]]
                if not ldef.genusIds:
                    continue
                if not ldef.restrictions:
                    continue
                genus_elts = [node_to_element_name[g] for g in ldef.genusIds]
                differentia = [
                    f"{adapter.label(r.propertyId)} some {self.node_to_name(r.fillerId)}"
                    for r in ldef.restrictions
                ]
                element.equivalent_to = (
                    f"{' and '.join(genus_elts)} and {' and '.join(differentia)}"
                )
                logging.info(f"Equiv[{element.name}] = {element.equivalent_to}")
        if ontology.elements is not None:
            for element in element_index.values():
                ontology.elements.append(element)
        return ontology

    def node_to_name(self, curie: str, label: Optional[str] = None) -> str:
        """Convert a node to a name.

        :param curie:
        :param label:
        :return:
        """
        if label is None:
            label = self.adapter.label(curie)
        if label is None:
            logger.warning(f"Node {curie} has no label")
            label = curie
        return camelcase(label)



"""Synonym engine."""
import logging
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Union

from jinja2 import Template
from oaklib import BasicOntologyInterface, get_adapter
from oaklib.datamodels.vocabulary import IS_A, SKOS_RELATED_MATCH
from oaklib.interfaces import MappingProviderInterface
from oaklib.types import CURIE
from pydantic import BaseModel
from sssom.parsers import parse_sssom_table, to_mapping_set_document
from sssom_schema import Mapping

from ontogpt.engines.knowledge_engine import KnowledgeEngine
from ontogpt.prompts.mapping import DEFAULT_MAPPING_EVAL_PROMPT

logger = logging.getLogger(__name__)

MAX_TOKENS = 300


class MappingPredicate(str, Enum):
    """Mapping predicates."""

    EXACT_MATCH = "exact_match"
    CLOSE_MATCH = "close_match"
    BROADER_THAN = "broader_than"
    NARROWER_THAN = "narrower_than"
    RELATED_TO = "related_to"
    DIFFERENT_FROM = "different_from"
    UNCATEGORIZED = "uncategorized"

    def mappings(self) -> Dict[str, str]:
        """Return the mappings for this predicate."""
        return {
            "skos:exactMatch": "exact_match",
            "skos:closeMatch": "close_match",
            "skos:broadMatch": "broader_than",
            "skos:narrowMatch": "narrower_than",
            "skos:relatedMatch": "related_to",
            "owl:differentFrom": "different",
        }


class Confidence(str, Enum):
    """Confidence levels."""

    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"


class CategorizedMapping(BaseModel):
    subject: CURIE = None
    object: CURIE = None
    completion: str = None
    predicate: Union[MappingPredicate, str] = None
    confidence: Union[Confidence, str] = None
    similarities: List[str] = None
    differences: List[str] = None


class MappingTask(BaseModel):
    subject: CURIE
    object: CURIE
    subject_label: str = None
    object_label: str = None
    subject_source: str = None
    object_source: str = None
    subject_adapter: str = None
    object_adapter: str = None
    predicate: Union[str, MappingPredicate] = None
    difficulty: Confidence = None
    score: float = None


class MappingTaskCollection(BaseModel):
    name: Optional[str] = None
    predicates: List[MappingPredicate] = []
    subjects: Optional[List[CURIE]] = None
    object_sources: Optional[List[str]] = None
    source_adapters: Optional[Dict[str, str]] = None
    tasks: List[MappingTask] = []


class Relationship(BaseModel):
    predicate: CURIE
    predicate_label: str
    object: CURIE
    object_label: str


class Concept(BaseModel):
    id: CURIE
    label: str
    definition: str = None
    synonyms: List[str] = []
    parents: List[str] = []
    relationships: List[Relationship] = []
    categories: List[str] = []


@dataclass
class MappingEngine(KnowledgeEngine):
    """Engine for generating and resolving mappings."""

    subject_adapter: BasicOntologyInterface = None
    object_adapter: BasicOntologyInterface = None

    def categorize_mapping(
        self, subject: CURIE, object: CURIE, template_path: str = None
    ) -> CategorizedMapping:
        if template_path is None:
            template_path = str(DEFAULT_MAPPING_EVAL_PROMPT)
        if len(template_path) == 0:
            template_path = str(DEFAULT_MAPPING_EVAL_PROMPT)
        if isinstance(template_path, Path):
            template_path = str(template_path)
        if isinstance(template_path, str):
            # create a Jinja2 template object
            with open(template_path) as file:
                template_txt = file.read()
                template = Template(template_txt)
        subject_concept = self._concept(subject, self.subject_adapter)
        object_concept = self._concept(object, self.object_adapter)
        prompt = template.render(
            subject=subject_concept,
            object=object_concept,
        )
        payload = self.client.complete(prompt, max_tokens=MAX_TOKENS)
        cm = self._parse(payload)
        cm.subject = str(subject)
        cm.object = str(object)
        if cm.predicate is None:
            cm.predicate = MappingPredicate.UNCATEGORIZED.value
        return cm

    def _parse(self, payload: str):
        cm = CategorizedMapping(
            completion=payload,
        )

        def _split(s: str, sep: str = ";") -> List[str]:
            if s is None:
                return []
            return [s.strip() for s in s.split(sep)]

        lines = payload.split("\n")
        for line in lines:
            if ":" not in line:
                logger.warning(f"Invalid line {line} in mapping response")
                continue
            key, value = line.split(":", 1)
            key = key.strip()
            value = value.strip()
            if key == "category":
                cm.predicate = value
                # try:
                #    cm.predicate = MappingPredicate[value]
                # except KeyError:
                #    cm.predicate = value
            elif key == "confidence":
                cm.confidence = value
                # try:
                #    cm.confidence = Confidence[value]
                # except KeyError:
                #    cm.confidence = value
            elif key == "similarities":
                cm.similarities = _split(value)
            elif key == "differences":
                cm.differences = _split(value)
            else:
                logger.warning(f"Unknown key {key} in mapping response")
        return cm

    def categorize_mappings(
        self, subjects: List[CURIE], object_sources: List[CURIE] = None
    ) -> Iterator[CategorizedMapping]:
        sa = self.subject_adapter
        if not isinstance(sa, MappingProviderInterface):
            raise TypeError(f"subject adapter {sa} must implement MappingProviderInterface")
        for object_source in object_sources:
            print(f"EVAL: {object_source} x {subjects}")
            for mapping in sa.sssom_mappings(subjects, source=object_source):
                print(f" MAP: {mapping}")
                yield self.categorize_mapping(mapping.subject_id, mapping.object_id)

    def run_tasks(
        self, task_collection: MappingTaskCollection, evaluate: bool = False
    ) -> Iterator[CategorizedMapping]:
        for task in task_collection.tasks:
            if task.subject_adapter:
                # TODO: do not mutate.
                self.subject_adapter = get_adapter(task.subject_adapter)
            if task.object_adapter:
                # TODO: do not mutate.
                self.object_adapter = get_adapter(task.object_adapter)
            cm = self.categorize_mapping(task.subject, task.object)
            yield cm

    def generate_tasks(
        self,
        task_collection: MappingTaskCollection = None,
        subjects: List[CURIE] = None,
        object_sources: List[CURIE] = None,
    ) -> Iterator[MappingTask]:
        sa = self.subject_adapter
        if task_collection:
            if task_collection.object_sources:
                if object_sources:
                    raise ValueError("object_sources specified twice")
                object_sources = task_collection.object_sources
            if task_collection.subjects:
                if subjects:
                    raise ValueError(
                        "subjects cannot be specified in both task_collection and as an argument"
                    )
                subjects = task_collection.subjects
        if not isinstance(sa, MappingProviderInterface):
            raise TypeError(f"subject adapter {sa} must implement MappingProviderInterface")
        for object_source in object_sources:
            for mapping in sa.sssom_mappings(subjects, source=object_source):
                yield MappingTask(
                    subject=mapping.subject_id,
                    object=mapping.object_id,
                    subject_source=mapping.subject_source,
                    object_source=mapping.object_source,
                    subject_adapter=f"sqlite:obo:{mapping.subject_source.lower()}",
                    object_adapter=f"sqlite:obo:{mapping.object_source.lower()}",
                )

    def from_sssom(
        self, path: Union[str, Path], exclude_functions: Optional[List[Callable]] = None
    ) -> MappingTaskCollection:
        """Generate tasks from an SSSOM file."""
        msdf = parse_sssom_table(path)
        msd = to_mapping_set_document(msdf)
        mappings = msd.mapping_set.mappings
        tasks = []
        for mapping in mappings:
            exclude = False
            if exclude_functions:
                for exclude_function in exclude_functions:
                    if exclude_function(mapping):
                        exclude = True
                        break
            if exclude:
                continue

            def _get_source(curie: CURIE) -> str:
                return curie.split(":")[0]

            mapping.subject_source = _get_source(mapping.subject_id)
            mapping.object_source = _get_source(mapping.object_id)
            mp = MappingPredicate()
            task = MappingTask(
                subject=mapping.subject_id,
                object=mapping.object_id,
                predicate=mp.mappings().get(mapping.predicate_id),
                subject_label=mapping.subject_label,
                object_label=mapping.object_label,
                subject_source=mapping.subject_source,
                object_source=mapping.object_source,
                subject_adapter=f"sqlite:obo:{mapping.subject_source.lower()}",
                object_adapter=f"sqlite:obo:{mapping.object_source.lower()}",
            )
            tasks.append(task)
        return MappingTaskCollection(
            tasks=tasks,
        )

    def categorize_sssom_mapping(
        self,
        mapping: Mapping,
    ) -> Tuple[Mapping, CategorizedMapping]:
        mapping = deepcopy(mapping)

        def _get_source(curie: CURIE) -> str:
            return curie.split(":")[0]

        def _get_adapter(src: str):
            return get_adapter(f"sqlite:obo:{src.lower()}")

        mapping.subject_source = _get_source(mapping.subject_id)
        mapping.object_source = _get_source(mapping.object_id)
        self.subject_adapter = _get_adapter(mapping.subject_source)
        self.object_adapter = _get_adapter(mapping.object_source)
        cm = self.categorize_mapping(mapping.subject_id, mapping.object_id)
        mp = MappingPredicate()
        revmap = {v.upper(): k for k, v in mp.mappings().items()}
        if cm.predicate.upper() not in revmap:
            logger.warning(f"Unknown predicate {cm.predicate}")
        mapping.predicate_id = revmap.get(cm.predicate.upper(), SKOS_RELATED_MATCH)
        return mapping, cm

    def _concept(self, curie: CURIE, adapter: BasicOntologyInterface) -> Concept:
        """Get a concept."""

        def _label(curie: CURIE) -> str:
            return adapter.label(curie)

        def _relationship(p, o) -> Relationship:
            return Relationship(
                predicate=p,
                predicate_label=_label(p),
                object=o,
                object_label=_label(o),
            )

        parents = [
            _label(rel[2])
            for rel in sorted(list(adapter.relationships([curie], predicates=[IS_A])))
        ]
        rels = [
            _relationship(p, o)
            for _, p, o in sorted(list(adapter.relationships([curie])))
            if p != IS_A
        ]
        label = adapter.label(curie)
        if not label:
            logger.warning(f"Could not find label for {curie} in {adapter}")
        return Concept(
            id=str(curie),
            label=label,
            synonyms=sorted(list(adapter.aliases_by_curie(curie))),
            definition=adapter.definition(curie),
            parents=parents,
            relationships=rels,
        )



"""Command line interface for ontogpt."""

import codecs
import json
import logging
import pickle
import sys
from copy import deepcopy
from dataclasses import dataclass
from io import BytesIO, TextIOWrapper
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import click
import jsonlines
import yaml

from litellm import get_model_cost_map
from oaklib import get_adapter
from oaklib.cli import query_terms_iterator
from oaklib.interfaces import OboGraphInterface
from oaklib.io.streaming_csv_writer import StreamingCsvWriter
from sssom.parsers import parse_sssom_table, to_mapping_set_document
from sssom.util import to_mapping_set_dataframe

import ontogpt.ontex.extractor as extractor
from ontogpt import DEFAULT_MODEL, __version__
from ontogpt.clients.llm_client import LLMClient
from ontogpt.clients.pubmed_client import PubmedClient
from ontogpt.clients.soup_client import SoupClient
from ontogpt.clients.wikipedia_client import WikipediaClient
from ontogpt.engines.embedding_similarity_engine import SimilarityEngine
from ontogpt.engines.generic_engine import GenericEngine, QuestionCollection
from ontogpt.engines.halo_engine import HALOEngine  # type: ignore
from ontogpt.engines.knowledge_engine import KnowledgeEngine
from ontogpt.engines.mapping_engine import MappingEngine
from ontogpt.engines.pheno_engine import PhenoEngine
from ontogpt.engines.reasoner_engine import ReasonerEngine
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.engines.synonym_engine import SynonymEngine
from ontogpt.evaluation.resolver import create_evaluator
from ontogpt.io.csv_wrapper import parse_yaml_predictions, write_graph, write_obj_as_csv
from ontogpt.io.html_exporter import HTMLExporter
from ontogpt.io.markdown_exporter import MarkdownExporter
from ontogpt.io.template_loader import get_template_details, get_template_path
from ontogpt.utils.multilingual import multilingual_analysis

__all__ = [
    "main",
]

from ontogpt.io.owl_exporter import OWLExporter
from ontogpt.io.rdf_exporter import RDFExporter
from ontogpt.io.yaml_wrapper import dump_minimal_yaml
from ontogpt.templates.core import ExtractionResult


@dataclass
class Settings:
    """Global command line settings."""

    cache_db: Optional[str] = None


settings = Settings()


def _as_text_writer(f):
    if isinstance(f, TextIOWrapper):
        return f
    else:
        return codecs.getwriter("utf-8")(f)


def write_extraction(
    results: ExtractionResult,
    output: BytesIO,
    output_format: str,
    knowledge_engine: KnowledgeEngine,
    template: str,
):
    """Write results of extraction to a given output stream."""
    # Check if this result contains anything writable first
    if results.extracted_object:
        exporter: Union[MarkdownExporter, HTMLExporter, RDFExporter, OWLExporter]

        if output_format not in ["pickle"]:
            output = _as_text_writer(output)

        if output_format == "pickle":
            output.write(pickle.dumps(results))
        elif output_format == "md":
            exporter = MarkdownExporter()
            exporter.export(results, output)
        elif output_format == "html":
            exporter = HTMLExporter(output=output)
            exporter.export(results, output)
        elif output_format == "yaml":
            output.write("---\n")  # type: ignore
            output.write(dump_minimal_yaml(results))  # type: ignore
        elif output_format == "turtle":
            exporter = RDFExporter()
            exporter.export(results, output, knowledge_engine.schemaview)
        elif output_format == "owl":
            exporter = OWLExporter()
            exporter.export(results, output, knowledge_engine.schemaview)
        elif output_format == "kgx":
            # TODO: enable passing name without extension,
            # since there will be multiple output files
            # TODO: rewrite to align with other exporters
            # by moving code into a dedicated KGXExporter class
            output.write("---\n")  # type: ignore
            output.write(dump_minimal_yaml(results))  # type: ignore

            # Need to return to the top of the output just written
            output.seek(0)

            if "." in template:
                module_name, class_name = template.split(".", 1)
            else:
                module_name = template
                class_name = None
            schema_path = get_template_path(module_name)
            # TODO: schema_path should really be a Path object
            nodes, edges = parse_yaml_predictions(
                yaml_path=output.name,  # type: ignore
                schema_path=schema_path,  # type: ignore
                root_class=class_name,  # type: ignore
            )
            nodestr, edgestr = write_graph(nodes, edges)
            with open("nodes.tsv", "w") as outfile:
                outfile.write(nodestr)
            with open("edges.tsv", "w") as outfile:
                outfile.write(edgestr)
        else:
            output.write("---\n")  # type: ignore
            output.write(dump_minimal_yaml(results))  # type: ignore


inputfile_option = click.option("-i", "--inputfile", help="Path to a file containing input text.")
template_option = click.option(
    "-t",
    "--template",
    required=True,
    help="Template to use. This may be the name of a predefined template"
    " or a path to a schema file. In the latter case, the schema file should"
    " be a YAML file and the path should include the .yaml file suffix.",
)
target_class_option = click.option(
    "-T", "--target-class", help="Target class (if not already root)."
)
model_option = click.option(
    "-m",
    "--model",
    help="Model name to use, e.g. orca-mini-7b or gpt-4."
    " See all model names with ontogpt list-models.",
)
prompt_template_option = click.option(
    "--prompt-template", help="Path to a file containing the prompt."
)
recurse_option = click.option(
    "--recurse/--no-recurse", default=True, show_default=True, help="Recursively parse structures."
)
use_textract_options = click.option(
    "--use-textract/--no-use-textract",
    default=False,
    show_default=True,
    help="Use textract to extract text.",
)
output_option_wb = click.option(
    "-o", "--output", type=click.File(mode="wb"), default=sys.stdout, help="Output file."
)
output_option_txt = click.option(
    "-o", "--output", type=click.File(mode="w"), default=sys.stdout, help="Output file."
)
output_format_options = click.option(
    "-O",
    "--output-format",
    type=click.Choice(["json", "yaml", "pickle", "md", "html", "owl", "turtle", "jsonl", "kgx"]),
    default="yaml",
    help="Output format.",
)
auto_prefix_option = click.option(
    "--auto-prefix",
    default="AUTO",
    help="Prefix to use for auto-generated classes. Default is AUTO.",
)
show_prompt_option = click.option(
    "--show-prompt/--no-show-prompt",
    default=False,
    show_default=True,
    help="If set, show all prompts passed to model through an API. Use with verbose setting.",
)
api_base_option = click.option(
    "--api-base",
    help="Base to use for LLM API, e.g. for the Azure OpenAI API."
    " Note this may also be set through the runoak set-apikey command.",
)
api_version_option = click.option(
    "--api-version",
    help="Version to use for LLM API, e.g. for the Azure OpenAI API."
    " Note this may also be set through the runoak set-apikey command.",
)
model_provider_option = click.option(
    "--model-provider",
    help="Specify a provider if model is not specified in the model name."
    " If using a proxy using the OpenAI API format, this should be set to 'openai'.",
)
temperature_option = click.option(
    "-p",
    "--temperature",
    type=click.FLOAT,
    default=1.0,
    help="Temperature for model completion.",
)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
@click.option("--cache-db", help="Path to sqlite database to cache prompt-completion results")
@click.version_option(__version__)
def main(verbose: int, quiet: bool, cache_db: str):
    """CLI for ontogpt.

    :param verbose: Verbosity while running.
    :param quiet: Boolean to be quiet or verbose.
    """
    logger = logging.getLogger()
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)
    logger.info(f"Logger {logger.name} set to level {logger.level}")
    if cache_db:
        settings.cache_db = cache_db


@main.command()
@inputfile_option
@template_option
@target_class_option
@model_option
@recurse_option
@output_option_wb
@click.option("--dictionary")
@output_format_options
@use_textract_options
@auto_prefix_option
@show_prompt_option
@click.option(
    "--set-slot-value",
    "-S",
    multiple=True,
    help="Set slot value, e.g. --set-slot-value has_participant=protein",
)
@click.argument("input", required=False)
@api_base_option
@api_version_option
@model_provider_option
@temperature_option
def extract(
    inputfile,
    template,
    target_class,
    dictionary,
    input,
    output,
    output_format,
    set_slot_value,
    use_textract,
    model,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Extract knowledge from text guided by schema, using SPIRES engine.

    Example:

        ontogpt extract -t gocam.GoCamAnnotations -i gocam-27929086.txt

    The input argument may be:
        A file path,
        A directory,
        or a string.
    Use the -i/--input-file option followed by the path to the input file
    or directory.
    If the input is a directory, all files with the .txt extension will be read.
    This is not recursive.
    Otherwise, the input is assumed to be a string to be read as input.

    You can also use fragments of existing schemas, use the --target-class option (-T) to
    specify an alternative Container/root class.

    Example:

        ontogpt extract -t gocam.GoCamAnnotations -T GeneOrganismRelationship "the mouse Shh gene"

    """
    # Choose model based on input, or use the default
    if not model:
        model = DEFAULT_MODEL

    inputlist = []

    if not inputfile or inputfile == "-":
        text = sys.stdin.read()
        inputlist.append(text)
    elif inputfile and Path(inputfile).is_dir():
        logging.info(f"Input file directory: {inputfile}")
        inputfiles = Path(inputfile).glob("*.txt")
        inputlist = [open(f, "r").read() for f in inputfiles if f.is_file()]
        logging.info(f"Found {len(inputlist)} input files here.")
    elif inputfile and Path(inputfile).exists():
        logging.info(f"Input file: {inputfile}")
        if use_textract:
            import textract

            text = textract.process(inputfile).decode("utf-8")
        else:
            text = open(inputfile, "r").read()
        logging.info(f"Input text: {text}")
        inputlist.append(text)
    elif inputfile and not Path(inputfile).exists():
        raise FileNotFoundError(f"Cannot find input file {inputfile}")

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db

    if dictionary:
        ke.load_dictionary(dictionary)
    if target_class:
        schemaview = template_details[3]
        target_class_def = schemaview.get_class(target_class)
    else:
        target_class_def = None

    i = 0
    for input_entry in inputlist:
        if len(inputlist) > 1:
            i = i + 1
            logging.info(f"Now extracting from file {i} of {len(inputlist)}")
        results = ke.extract_from_text(
            text=input_entry, cls=target_class_def, show_prompt=show_prompt
        )
        if set_slot_value:
            for slot_value in set_slot_value:
                slot, value = slot_value.split("=")
                setattr(results.extracted_object, slot, value)
        write_extraction(results, output, output_format, ke, template)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@auto_prefix_option
@show_prompt_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("entity")
def generate_extract(
    model,
    entity,
    template,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Generate text and then extract knowledge from it.
    
    Example:

    ontogpt generate-extract -m ollama/llama3 -t kidney "kidney failure"
    """
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db

    logging.debug(f"Input entity: {entity}")
    results = ke.generate_and_extract(
        entity=entity, prompt_template=template, show_prompt=show_prompt
    )
    write_extraction(results, output, output_format, ke, template)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@auto_prefix_option
@show_prompt_option
@click.option("--ontology", "-r", help="Ontology to use; use oaklib selector path")
@click.option("--max-iterations", "-M", default=10, type=click.INT)
@click.option("--iteration-slot", "-I", multiple=True, help="Slots to iterate over")
@click.option("--db", "-D", help="Where the resulting yaml database is stored")
@click.option(
    "--clear/--no-clear", default=False, show_default=True, help="Clear the db before starting"
)
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("entity")
def iteratively_generate_extract(
    model,
    entity,
    template,
    output,
    output_format,
    db,
    iteration_slot,
    max_iterations,
    clear,
    ontology,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Iterate through generate-extract."""
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db

    logging.debug(f"Input entity: {entity}")
    adapter = get_adapter(ontology)
    for results in ke.iteratively_generate_and_extract(
        entity,
        db,
        show_prompt=show_prompt,
        iteration_slots=list(iteration_slot),
        max_iterations=max_iterations,
        adapter=adapter,
        clear=clear,
    ):
        write_extraction(results, output, output_format, ke, template)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@show_prompt_option
@click.option(
    "--limit",
    default=20,
    help="Total number of citation records to return.",
)
@click.option(
    "--get-pmc/--no-get-pmc",
    default=False,
    help="Attempt to parse PubMed Central full text(s) instead of abstract(s) alone.",
)
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("search")
@click.option(
    "--max-text-length",
    default=3000,
    help="Maximum text length for each input chunk. Dependent on context size of model used.",
)
@click.argument("search")
def pubmed_annotate(
    model,
    search,
    template,
    output,
    output_format,
    limit,
    get_pmc,
    show_prompt,
    max_text_length,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Retrieve a collection of PubMed IDs for a search term; annotate them using a template.

    Example:
    ontogpt pubmed-annotate -t phenotype "Takotsubo Cardiomyopathy: A Brief Review"
        --get-pmc --model gpt-3.5-turbo-16k --limit 3
    """
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db

    pubmed_annotate_limit = limit
    pmc = PubmedClient()
    pmc.max_text_length = max_text_length
    pmids = pmc.get_pmids(search)
    if get_pmc:
        logging.info("Will try to retrieve PubMed Central texts.")
        textlist = pmc.text(pmids[: pubmed_annotate_limit + 1], pubmedcental=True)
    else:
        textlist = pmc.text(pmids[: pubmed_annotate_limit + 1])
    for text in textlist:
        logging.debug(f"Input text: {text}")
        results = ke.extract_from_text(text=text, show_prompt=show_prompt)
        write_extraction(results, output, output_format, ke, template)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@show_prompt_option
@click.option("--auto-prefix", default="AUTO", help="Prefix to use for auto-generated classes.")
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("article")
def wikipedia_extract(
    model,
    article,
    template,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Extract knowledge from a Wikipedia page."""
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db

    logging.info(f"Creating for {template} => {article}")
    client = WikipediaClient()
    text = client.text(article)

    logging.debug(f"Input text: {text}")
    results = ke.extract_from_text(text=text, show_prompt=show_prompt)
    write_extraction(results, output, output_format, ke, template)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@show_prompt_option
@click.option(
    "--keyword",
    "-k",
    multiple=True,
    help="Keyword to search for (e.g. --keyword therapy). Also obtained from schema",
)
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("topic")
def wikipedia_search(
    model,
    topic,
    keyword,
    template,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Extract knowledge from a Wikipedia page."""
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )

    logging.info(f"Creating for {template} => {topic}")
    client = WikipediaClient()
    keywords = list(keyword) if keyword else []
    logging.info(f"KW={keywords}")

    keywords.extend(ke.schemaview.schema.keywords)
    search_term = f"{topic + ' ' + ' '.join(keywords)}"
    print(f"Searching for {search_term}")
    search_results = client.search_wikipedia_articles(search_term)
    for _index, result in enumerate(search_results, start=1):
        title = result["title"]
        text = client.text(title)
        logging.debug(f"Input text: {text}")
        results = ke.extract_from_text(text=text, show_prompt=show_prompt)
        write_extraction(results, output, output_format, ke, template)
        break


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@show_prompt_option
@click.option(
    "--keyword",
    "-k",
    multiple=True,
    help="Keyword to search for (e.g. --keyword therapy). Also obtained from schema",
)
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("term_tokens", nargs=-1)
def search_and_extract(
    model,
    term_tokens,
    keyword,
    template,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Search for relevant literature through PubMed and extract knowledge from it.
    
    Example:

    ontogpt search-and-extract -m ollama/llama3 -t phenotype --keyword "kidney failure"
    """
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )

    term = " ".join(term_tokens)
    logging.info(f"Creating for {template}; search={term} kw={keyword}")

    logging.info(f"Creating PubMed client for {template}; search={term}")
    pmc = PubmedClient()
    logging.info("Got client")
    keywords = list(keyword) if keyword else []
    logging.info(f"KW={keywords}")
    keywords.extend(ke.schemaview.schema.keywords)
    logging.info(f"Keywords={keywords}")
    if not keywords:
        raise ValueError("No keywords specified; use --keyword or annotate schema with keywords")
    pmids = list(pmc.search(term, keywords))
    logging.info(f"PMIDs={pmids}")
    pmid = pmids[0]
    logging.info(f"PMID={pmid}")
    text = pmc.text(pmid)
    logging.info(f"Input text: {text}")
    results = ke.extract_from_text(text=text, show_prompt=show_prompt)
    write_extraction(results, output, output_format, ke, template)


@main.command()
@template_option
@model_option
@recurse_option
@output_option_wb
@output_format_options
@show_prompt_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("url")
def web_extract(
    model,
    template,
    url,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Extract knowledge from web page."""
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db

    web_client = SoupClient()
    text = web_client.text(url)

    logging.debug(f"Input text: {text}")
    results = ke.extract_from_text(text=text, show_prompt=show_prompt)
    write_extraction(results, output, output_format, ke, template)


@main.command()
@output_option_wb
@click.option("--dictionary")
@output_format_options
@click.option(
    "--recipes-urls-file",
    "-R",
    help="File with URLs to recipes to use for extraction",
)
@click.option("--auto-prefix", default="AUTO", help="Prefix to use for auto-generated classes.")
@model_option
@show_prompt_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("url")
def recipe_extract(
    model,
    url,
    recipes_urls_file,
    dictionary,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Extract from recipe on the web."""
    try:
        from recipe_scrapers import scrape_me
    except ModuleNotFoundError as e:
        logging.error(
            f"Did not find recipe_scrapers. Try: poetry install extras=recipes. Error: {e}"
        )

    if not model:
        model = DEFAULT_MODEL

    template_details = get_template_details(template="recipe")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    if settings.cache_db:
        ke.client.cache_db_path = settings.cache_db

    if recipes_urls_file:
        with open(recipes_urls_file, "r") as f:
            urls = [line.strip() for line in f.readlines() if url in line]
            if len(urls) != 1:
                raise ValueError(f"Found {len(urls)} URLs in {recipes_urls_file}")
            url = urls[0]
    scraper = scrape_me(url)

    if dictionary:
        ke.load_dictionary(dictionary)
    ingredients = "\n".join(scraper.ingredients())
    instructions = "\n".join(scraper.instructions_list())
    text = f"""
    Recipe: {scraper.title()}
    Ingredients:\n{ingredients}
    Instructions:\n{instructions}
    """
    logging.info(f"Input text: {text}")
    results = ke.extract_from_text(text=text, show_prompt=show_prompt)
    logging.debug(f"Results: {results}")
    results.extracted_object.url = url
    write_extraction(results, output, output_format, ke, "recipe")


@main.command()
@model_option
@template_option
@output_option_wb
@output_format_options
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("input")
def convert(
    model,
    template,
    input,
    output,
    output_format,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Convert output format."""
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )

    cls = ke.template_pyclass
    with open(input, "r") as f:
        data = yaml.safe_load(f)
    obj = cls(**data["extracted_object"])
    results = ExtractionResult(extracted_object=obj)
    write_extraction(results, output, output_format, ke, template)


@main.command()
@model_option
@output_option_txt
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.option(
    "-C", "--context", required=True, help="domain e.g. anatomy, industry, health-related"
)
@click.argument("term")
def synonyms(
    model, term, context, output, temperature, api_base, api_version, model_provider, **kwargs
):
    """Extract synonyms.
    
    Examples:

    ontogpt synonyms -m ollama/llama3 --context "political" "abdicate"

    ontogpt synonyms -m ollama/llama3 --context "biological" "dessicate"

    """
    logging.info(f"Creating for {term}")

    if not model:
        model = DEFAULT_MODEL

    ke = SynonymEngine(
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    out = ke.synonyms(term, context)
    for line in out:
        output.write(f"{line}\n")


@main.command()
@model_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("text", nargs=-1)
def embed(text, model, api_base, api_version, model_provider, **kwargs):
    """Embed text."""
    if model is None:
        model = "text-embedding-ada-002"
    logging.info(f"Using model {model} for embeddings.")

    logging.info(f"Embedding with model {model}")

    if not text:
        raise ValueError("Text must be passed to this function.")

    client = LLMClient(
        model=model, api_base=api_base, api_version=api_version, custom_llm_provider=model_provider
    )
    resp = client.embeddings(text)
    print(resp)


@main.command()
@model_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("text", nargs=-1)
def text_similarity(text, model, api_base, api_version, model_provider, **kwargs):
    """Get similarity between two text inputs.

    Text should be separated by @, e.g., "text1 @ text2".
    """
    if model is None:
        model = "text-embedding-ada-002"
    logging.info(f"Using model {model} for embeddings.")

    logging.info(f"Embedding with model {model}")

    if not text:
        raise ValueError("Text must be passed to this function.")
    text = list(text)
    if "@" not in text:
        raise ValueError("Texts must be separated with @")
    ix = text.index("@")
    text1 = " ".join(text[:ix])
    text2 = " ".join(text[ix + 1 :])
    logging.info(text1)
    logging.info(text2)

    client = LLMClient(
        model=model, api_base=api_base, api_version=api_version, custom_llm_provider=model_provider
    )
    sim = client.similarity(text1, text2)
    print(sim)


@main.command()
@model_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("text", nargs=-1)
def text_distance(text, model, api_base, api_version, model_provider, **kwargs):
    """Embed text and calculate euclidian distance between embeddings.

    Text should be separated by @, e.g., "text1 @ text2".
    """
    if model is None:
        model = "text-embedding-ada-002"
    logging.info(f"Using model {model} for embeddings.")

    logging.info(f"Embedding with model {model}")

    if not text:
        raise ValueError("Text must be passed to this function.")
    text = list(text)
    if "@" not in text:
        raise ValueError("Text must be separated with @")
    ix = text.index("@")
    text1 = " ".join(text[:ix])
    text2 = " ".join(text[ix + 1 :])
    logging.info(text1)
    logging.info(text2)

    client = LLMClient(
        model=model, api_base=api_base, api_version=api_version, custom_llm_provider=model_provider
    )
    sim = client.euclidian_distance(text1, text2)
    print(sim)


@main.command()
@output_option_txt
@output_format_options
@model_option
@click.option("--ontology", "-r", help="Ontology to use")
@click.option(
    "--definitions/--no-definitions",
    default=True,
    show_default=True,
    help="Include text definitions in the text to embed",
)
@click.option(
    "--parents/--no-parents",
    default=True,
    show_default=True,
    help="Include is-a parent terms in the text to embed",
)
@click.option(
    "--ancestors/--no-ancestors",
    default=True,
    show_default=True,
    help="Include all ancestors in the text to embed",
)
@click.option(
    "--logical-definitions/--no-logical-definitions",
    default=True,
    show_default=True,
    help="Include logical definitions in the text to embed",
)
@click.option(
    "--autolabel/--no-autolabel",
    default=True,
    show_default=True,
    help="Add subj/obj labels to report objects",
)
@click.option(
    "--synonyms/--no-synonyms",
    default=True,
    show_default=True,
    help="Include synonyms in the text to embed",
)
@click.argument("terms", nargs=-1)
@api_base_option
@api_version_option
@model_provider_option
def entity_similarity(
    terms, ontology, output, model, output_format, api_base, api_version, model_provider, **kwargs
):
    """Identify similarity between two entities in an ontology.

    Text should be separated by @, e.g., "entitiy1 @ entity2".
    """

    if not terms:
        raise ValueError("terms must be passed")
    terms = list(terms)
    if "@" not in terms:
        logging.info("No @ found, assuming all by all")
        terms1 = list(terms)
        terms2 = list(terms)
    else:
        ix = terms.index("@")
        terms1 = terms[:ix]
        terms2 = terms[ix + 1 :]
    adapter = get_adapter(ontology)
    entities1 = list(query_terms_iterator(terms1, adapter))
    entities2 = list(query_terms_iterator(terms2, adapter))

    engine = SimilarityEngine(
        model=model,
        adapter=adapter,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )
    writer = StreamingCsvWriter(output, heterogeneous_keys=False)

    for e1 in entities1:
        sims = engine.search(e1, entities2)
        for sim in sims:
            writer.emit(sim)


@main.command()
@inputfile_option
@output_option_txt
@model_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.option("--task-file")
@click.option("--task-type")
@click.option("--tsv-output")
@click.option("--all-methods/--no-all-methods", default=False)
@click.option("--explain/--no-explain", default=False)
@click.option("--evaluate/--no-evaluate", default=False)
@click.argument("terms", nargs=-1)
def reason(
    terms,
    inputfile,
    model,
    task_file,
    explain,
    task_type,
    output,
    tsv_output,
    all_methods,
    evaluate,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Reason."""
    reasoner = ReasonerEngine(
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
    )
    if task_file:
        tc = extractor.TaskCollection.load(task_file)
    else:
        adapter = get_adapter(inputfile)
        if not isinstance(adapter, OboGraphInterface):
            raise ValueError("Only OBO graphs supported")
        ex = extractor.OntologyExtractor(adapter=adapter)
        # ex.use_identifiers = True
        task = ex.create_task(task_type=task_type, parameters=list(terms))
        tc = extractor.TaskCollection(tasks=[task])
    if all_methods:
        tasks = []
        print(f"Cloning {len(tc.tasks)} tasks")
        for core_task in tc.tasks:
            for m in extractor.LLMReasonMethodType:
                print(f"Cloning {m}")
                task = deepcopy(core_task)
                task.method = m
                task.init_method()
                tasks.append(task)
        tc.tasks = tasks
        print(f"New {len(tc.tasks)} tasks")
    else:
        for task in tc.tasks:
            task.include_explanations = explain
    resultset = reasoner.reason_multiple(tc, evaluate=evaluate)
    dump_minimal_yaml(resultset.model_dump(), file=output)
    if tsv_output:
        write_obj_as_csv(resultset.results, tsv_output)


@main.command()
@output_option_txt
@model_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("phenopacket_files", nargs=-1)
def diagnose(
    phenopacket_files,
    model,
    output,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Diagnose a clinical case represented as one or more Phenopackets."""
    if not phenopacket_files:
        raise ValueError("No phenopacket files specified. Please provide one or more files.")

    if not model:
        model = DEFAULT_MODEL

    phenopackets = [json.load(open(f)) for f in phenopacket_files]
    engine = PhenoEngine(
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
    )
    results = engine.evaluate(phenopackets)
    print(dump_minimal_yaml(results))
    write_obj_as_csv(results, output)


@main.command()
@click.argument("input_data_dir")
@click.argument("output_directory")
@output_option_wb
@model_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
def run_multilingual_analysis(
    input_data_dir,
    output_directory,
    output,
    temperature,
    api_base,
    api_version,
    model_provider,
    model="gpt-4-turbo",
):
    """Call the multilingual analysis function."""
    multilingual_analysis(
        input_data_dir=input_data_dir, output_directory=output_directory, output=output, model=model
    )


@main.command()
@inputfile_option
@output_option_txt
@model_option
@click.option("--tsv-output")
@click.option("--template-path")
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
def answer(
    inputfile,
    model,
    template_path,
    output,
    tsv_output,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Answer a set of questions defined in YAML."""
    qc = QuestionCollection(**yaml.safe_load(open(inputfile)))
    engine = GenericEngine(
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
    )
    qs = []
    for q in engine.run(qc, template_path=template_path):
        print(dump_minimal_yaml(q))
        qs.append(q)
    qc.questions = qs
    output.write(dump_minimal_yaml(qs))
    if tsv_output:
        write_obj_as_csv(qs, tsv_output)


@main.command()
@inputfile_option
@output_option_txt
@model_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.option("--task-file")
@click.option("--task-type")
@click.option("--tsv-output")
@click.option("--yaml-output")
@click.option("--all-methods/--no-all-methods", default=False)
@click.option("--explain/--no-explain", default=False)
@click.option("--evaluate/--no-evaluate", default=False)
def categorize_mappings(
    inputfile,
    model,
    task_file,
    explain,
    task_type,
    output,
    tsv_output,
    yaml_output,
    all_methods,
    evaluate,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Categorize a collection of SSSOM mappings."""
    mapper = MappingEngine(
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
    )
    if tsv_output:
        tc = mapper.from_sssom(inputfile)
        for cm in mapper.run_tasks(tc, evaluate=evaluate):
            print(dump_minimal_yaml(cm.model_dump()))
            # dump_minimal_yaml(cm.dict(), file=output)
        # write_obj_as_csv(resultset.results, tsv_output)
    else:
        import sssom.writers as sssom_writers

        msdf = parse_sssom_table(inputfile)
        msd = to_mapping_set_document(msdf)
        mappings = []
        cms = []
        done = []
        for mapping in msd.mapping_set.mappings:
            pair = mapping.subject_id, mapping.object_id
            if pair in done:
                continue
            mapping, cm = mapper.categorize_sssom_mapping(mapping)
            mappings.append(mapping)
            cms.append(cm.model_dump())
            done.append(pair)
        msd.mapping_set.mappings = mappings
        msdf = to_mapping_set_dataframe(msd)
        sssom_writers.write_table(msdf, output)
        if yaml_output:
            with open(yaml_output, "w") as file:
                dump_minimal_yaml(cms, file=file)


@main.command()
@recurse_option
@model_option
@output_option_txt
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.option(
    "--num-tests",
    type=click.INT,
    default=5,
    show_default=True,
    help="number of iterations to cycle through.",
)
@click.option(
    "--chunking/--no-chunking",
    default=False,
    show_default=True,
    help="If set, chunk input text, then prepare a separate prompt for each chunk."
    " Otherwise the full input text is passed.",
)
@click.argument("evaluator")
def eval(
    evaluator,
    num_tests,
    output,
    chunking,
    model,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Evaluate an extractor."""
    logging.info(f"Creating for {evaluator}")

    evaluator = create_evaluator(
        name=evaluator, num_tests=num_tests, chunking=chunking, model=model
    )
    eos = evaluator.eval()
    output.write(dump_minimal_yaml(eos, minimize=False))


@main.command()
@template_option
@model_option
@click.option("-E", "--examples", type=click.File("r"), help="File of example objects.")
@recurse_option
@output_option_wb
@output_format_options
@show_prompt_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("object")
def fill(
    model,
    template,
    object: str,
    examples,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Fill in missing values, given examples."""
    ke: KnowledgeEngine

    # Choose model based on input, or use the default
    if not model:
        model = DEFAULT_MODEL

    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        **kwargs,
    )

    object = yaml.safe_load(object)
    logging.info(f"Object to fill =  {object}")
    logging.info(f"Loading {examples}")
    examples = yaml.safe_load(examples)
    logging.debug(f"Input object: {object}")
    results = ke.generalize(object=object, examples=examples, show_prompt=show_prompt)

    output.write(yaml.dump(results.model_dump()))


@main.command()
@inputfile_option
@model_option
@output_option_txt
@output_format_options
@show_prompt_option
@api_base_option
@api_version_option
@model_provider_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("input", required=False)
def complete(
    inputfile,
    model,
    input,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Prompt completion.

    The input argument may be:
        A file path,
        or a string.
    Use the -i/--input-file option followed by the path to the input file.
    Otherwise, the input is assumed to be a string to be read as input.
    """

    if inputfile:
        text = open(inputfile).read()
    else:
        text = input.strip()

    results = _send_complete_request(
        model,
        text,
        output,
        output_format,
        show_prompt,
        temperature,
        api_base,
        api_version,
        model_provider,
    )

    output.write(results + "\n")


def _send_complete_request(
    model,
    input,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
) -> str:
    """Send a completion request to an LLM endpoint."""

    if not model:
        model = DEFAULT_MODEL

    c = LLMClient(
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        custom_llm_provider=model_provider,
    )
    results = c.complete(prompt=input, show_prompt=show_prompt)

    return results


@main.command()
@template_option
@click.option("--input", "-i", type=click.File("r"), default=sys.stdin, help="Input file")
def parse(template, input):
    """Parse OpenAI results."""
    if template:
        template_details = get_template_details(template=template)
    else:
        raise ValueError("No template specified. Use -t/--template option.")

    ke = SPIRESEngine(template_details=template_details)
    text = input.read()
    logging.debug(f"Input text: {text}")
    # ke.annotator = BioPortalImplementation()
    results = ke.parse_completion_payload(text)
    print(yaml.dump(results))


# TODO: rewrite for use with litellm's disk cache
# See https://grantjenks.com/docs/diskcache/tutorial.html
@main.command()
@click.option("-o", "--output", type=click.File(mode="w"), default=sys.stdout, help="Output file.")
@output_format_options
@model_option
@click.option("-m", "match", help="Match string to use for filtering.")
@click.option("-D", "database", help="Path to sqlite database.")
def dump_completions(model, match, database, output, output_format):
    """Dump cached completions."""
    raise NotImplementedError("This function is not implemented at this time.")


@main.command()
@click.option("-o", "--output", type=click.File(mode="w"), default=sys.stdout, help="Output file.")
@click.argument("input", type=click.File("r"))
def convert_examples(input, output):
    """Convert training examples from YAML."""
    logging.info(f"Creating examples for {input}")
    example_doc = yaml.safe_load(input)
    writer = jsonlines.Writer(output)
    for example in example_doc["examples"]:
        prompt = example["prompt"]
        completion = yaml.dump(example["completion"], sort_keys=False)
        writer.write(dict(prompt=prompt, completion=completion))


@main.command()
@model_option
@click.option("-o", "--output", type=click.File(mode="w"), default=sys.stdout, help="Output file.")
@click.option("-i", "--input", help="Input ontology.")
@click.option("-c", "--context", help="Context.")
@click.option(
    "--num-iterations",
    type=click.INT,
    default=5,
    show_default=True,
    help="number of iterations to cycle through.",
)
@click.argument("terms", nargs=-1)
def halo(model, input, context, terms, output, **kwargs):
    """Run HALO over inputs."""

    engine = HALOEngine()
    engine.seed_from_file(input)
    if context is None:
        context = engine.ontology.elements[0].context
    engine.fixed_slot_values = {"context": context}
    engine.hallucinate(terms, **kwargs)
    output.write(dump_minimal_yaml(engine.ontology))


@main.command()
@model_option
@output_option_wb
@output_format_options
@show_prompt_option
@click.option(
    "-d",
    "--description",
    help="Patient description, in free text",
)
@click.option(
    "--sections", multiple=True, help="sections to include e.g. medications, vital signs, etc."
)
@api_base_option
@api_version_option
@model_provider_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
def clinical_notes(
    description,
    sections,
    output,
    model,
    show_prompt,
    output_format,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Create mock clinical notes.

    Example:

        ontogpt clinical-notes -d "middle-aged female patient with diabetes"
        ontogpt clinical-notes --description "middle-aged female patient with diabetes"\
         --sections medications --sections "vital signs"

    """
    prompt = "create mock clinical notes for a patient like this: " + description
    if sections:
        prompt += " including sections: " + ", ".join(sections)

    if not model:
        model = DEFAULT_MODEL

    c = LLMClient(
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        custom_llm_provider=model_provider,
    )
    results = c.complete(prompt=prompt, show_prompt=show_prompt)

    output.write(results)


@main.command()
def list_templates():
    """List the available extraction templates.

    The following values are provided:

    ID: The ID of the template. Use this with the -t/--template option in another command.

    Name: The name of the template.

    Description: A brief description of the template.
    
    """

    # Get the list of yaml files in the templates directory
    all_templates = _get_templates()

    # Sort that dict by id
    all_templates = dict(sorted(all_templates.items()))

    # Write it out
    print("ID\tName\tDescription")
    for template_id, (name, description) in all_templates.items():
        print(f"{template_id}\t{name}\t{description}")


def _get_templates() -> Dict[str, Tuple[str, str]]:
    """Retrieve the list of all templates."""

    http_prefixes = ("http", "https")

    # Get the list of yaml files in the templates directory
    # and populate a dict
    all_templates = {}
    template_dir = Path(__file__).parent / "templates"
    template_paths = [f for f in template_dir.glob("*.yaml")]
    for template_path in template_paths:
        with open(template_path, "r") as template_file:
            data = yaml.safe_load(template_file)
            if data["id"].startswith(http_prefixes):
                identifier = data["id"].split("/")[-1]
            else:
                identifier = data["id"]
            all_templates[identifier] = (data["name"], data["description"])

    return all_templates


@main.command()
def list_models():
    """List all available models.

    Note this is a partial list of models available through litellm
    for use in the OntoGPT CLI. More models may be available from
    other sources.

    The following values are provided:

    Model Name: Name of the model. Use this with the -m/--model option.

    Provider: The service provider for the model.

    Functionality: The relevance of the model to OntoGPT functions.
    "chat" or "completion" models are used for generating text and may
    be used with extract-based functions. "embedding" models are used
    for generating embeddings and may be used with similarity functions.

    Max Tokens: Token limit for the model. Note that models may
    tokenize text differently and calculate input and/or output tokens
    in particular ways, so consult a model's original documentaion for
    further details.
    """
    models = get_model_cost_map("")

    print("Model Name\tProvider\tFunctionality\tMax Tokens")
    for model in models:
        primary_name = model
        provider = models[model]["litellm_provider"]

        if "mode" in models[model]:
            functionality = models[model]["mode"]
            if functionality not in ["chat", "completion", "embedding"]:
                continue
        else:
            continue

        max_tokens = models[model]["max_tokens"]

        print(f"{primary_name}\t{provider}\t{functionality}\t{max_tokens}")


@main.command()
@model_option
@output_option_txt
@output_format_options
@show_prompt_option
@temperature_option
@api_base_option
@api_version_option
@model_provider_option
@click.argument("input")
def suggest_templates(
    input,
    model,
    output,
    output_format,
    show_prompt,
    temperature,
    api_base,
    api_version,
    model_provider,
    **kwargs,
):
    """Provide a suggestion for an appropriate template, given a text input.

    This is powered by the specified LLM.

    Example:

    ontogpt suggest-template "horses"
    ontogpt suggest-template "Takotsubo Cardiomyopathy"
    ontogpt suggest-template "I need to extract ingredients from recipes"

    """

    # Get the input text and preprocess it a bit
    input_text = (
        "Given the following table of templates, "
        f"which are most appropriate for the following topic: {input.strip()}"
    )

    # Get the list of templates and sort
    all_templates = _get_templates()
    all_templates = dict(sorted(all_templates.items()))

    # Assemble it into a string
    all_templates_string = "\n".join(
        [
            f"{template_id}\t{name}\t{description}"
            for template_id, (name, description) in all_templates.items()
        ]
    )

    input_text = input_text + "\n" + "ID\tName\tDescription\n" + all_templates_string

    # Use the complete function directly to address query
    results = _send_complete_request(
        model=model,
        input=input_text,
        output=output,
        output_format=output_format,
        show_prompt=show_prompt,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
    )

    output.write(results + "\n")


if __name__ == "__main__":
    main()



"""Loader functions for extraction templates."""

import importlib
import logging
from pathlib import Path

from linkml.generators.pydanticgen import PydanticGenerator
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import ClassDefinition

TEMPLATE_NAME = str

this_path = Path(__file__).parent
logger = logging.getLogger(__name__)


def get_template_details(template: TEMPLATE_NAME) -> ClassDefinition:
    """
    Get the LinkML class, module, and schemaview for a template.

    This may be a template already known to ontogpt or the path
    to a custom template. It is assumed to be the latter if the
    template input is a path to a YAML file.
    :param template: template name of the form module.ClassName
    :return: tuple of (LinkML class definition, module, specific
    Python class, schemaview)
    """
    # Check if template is a path to a YAML file
    # If so, this is a custom schema and needs python classes
    if template.endswith(".yaml"):
        logger.info(f"Loading custom schema from {template}")
        path_to_template = Path(template)

        # Check if file exists first
        if not path_to_template.exists():
            raise FileNotFoundError(f"Template file not found at {template}")

        # Copy the schema to the templates directory
        # So it will have access to imports like the core schema
        templates_path = this_path.parent / "templates"
        new_path_to_template = templates_path / path_to_template.name
        new_path_to_template.write_text(path_to_template.read_text())
        module_name = new_path_to_template.stem
        path_to_module = new_path_to_template.with_suffix(".py")

        sv = SchemaView(new_path_to_template)

        gen = PydanticGenerator(str(new_path_to_template), pydantic_version=2)
        path_to_module.write_text(gen.serialize())

        mod = importlib.import_module(f"ontogpt.templates.{module_name}")
        class_name = None

    else:
        logger.info(f"Loading schema for {template}")
        if "." in template:
            module_name, class_name = template.split(".", 1)
        else:
            module_name = template
            class_name = None
        path_to_template = get_template_path(module_name)
        sv = SchemaView(path_to_template)
        mod = importlib.import_module(f"ontogpt.templates.{module_name}")

    if class_name is None:
        roots = [c.name for c in sv.all_classes().values() if c.tree_root]
        if len(roots) != 1:
            raise ValueError(f"Template {template} does not have singular root: {roots}")
        class_name = roots[0]

    pyclass = mod.__dict__[class_name]

    logger.info(f"Getting class for template {template}")
    cls = None
    for c in sv.all_classes().values():
        if c.name == class_name:
            cls = c
            break
    if not cls:
        raise ValueError(f"Template {template} not found")
    return (cls, mod, pyclass, sv)


def get_template_path(module_name: str) -> Path:
    """Get the path to the YAML file for the template."""
    templates_path = this_path.parent / "templates"
    return templates_path / f"{module_name}.yaml"



from __future__ import annotations 
from datetime import (
    datetime,
    date
)
from decimal import Decimal 
from enum import Enum 
import re
import sys
from typing import (
    Any,
    List,
    Literal,
    Dict,
    Optional,
    Union
)
from pydantic.version import VERSION  as PYDANTIC_VERSION 
if int(PYDANTIC_VERSION[0])>=2:
    from pydantic import (
        BaseModel,
        ConfigDict,
        Field,
        field_validator
    )
else:
    from pydantic import (
        BaseModel,
        Field,
        validator
    )

metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment = True,
        validate_default = True,
        extra = "forbid",
        arbitrary_types_allowed = True,
        use_enum_values = True,
        strict = False,
    )
    pass


class NullDataOptions(str, Enum):
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    NOT_MENTIONED = "NOT_MENTIONED"


class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")


class NamedEntity(ConfiguredBaseModel):
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class CompoundExpression(ConfiguredBaseModel):
    pass


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")


class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)


class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    publication: Optional[Publication] = Field(None)
    entities: Optional[List[str]] = Field(default_factory=list)


class RelationshipType(NamedEntity):
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Publication(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")


class AnnotatorResult(ConfiguredBaseModel):
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)


class Container(ConfiguredBaseModel):
    interactions: Optional[List[BioticInteraction]] = Field(default_factory=list)


class BioticInteraction(ConfiguredBaseModel):
    source_taxon: Optional[str] = Field(None, description="""the taxon that is the subject of the interaction""")
    target_taxon: Optional[str] = Field(None, description="""the taxon that is the object of the interaction""")
    interaction_type: Optional[str] = Field(None, description="""the type of interaction""")


class Taxon(NamedEntity):
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class InteractionType(NamedEntity):
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
Container.model_rebuild()
BioticInteraction.model_rebuild()
Taxon.model_rebuild()
InteractionType.model_rebuild()

